#!/usr/bin/env python
# -*- coding: utf-8 -*-

## MoonShineLua (MSL)
##
## Copyright 2016 Andrei Costin (andrei@firmware.re) for Firmware.RE
## with support from Tjaldur Software Governance Solutions and NLnet.nl.
##
## See LICENSE file for details.


'''
Taint-related logic, utilities and (sub-)tree/node 
walking/analysis/annotation for MSL.
'''

import pprint

import antlr4
from antlr4 import *

from msl.parsing.lua.LuaLexer import LuaLexer
from msl.parsing.lua.LuaParser import LuaParser

import msl.utils.utils as mslutils
import msl.config.common as commonconfig

from treelib import Tree, Node
from treelib.tree import NodeIDAbsentError

def solve_tainting(cst_tlib, c, c_text, variables_to_expressions_map):
    variables_to_expressions_map_by_name = dict([itm[0], itm] for itm in variables_to_expressions_map.values())
    c_node_id = mslutils.get_object_id(c)
    c_tree_node = cst_tlib[c_node_id]
    c_tree_parent = cst_tlib.parent(c_node_id)
    c_tree_parent_parent = cst_tlib.parent(c_tree_parent.identifier)
    subtree_copy = None
    c_tree_parent_parent_identifier = None
    if c_tree_parent_parent == None:
        subtree_copy = Tree(cst_tlib.subtree(c_tree_parent.identifier), deep = False)
        c_tree_parent_parent_identifier = c_tree_parent.identifier
    else:
        subtree_copy = Tree(cst_tlib.subtree(c_tree_parent_parent.identifier), deep = False)
        c_tree_parent_parent_identifier = c_tree_parent_parent.identifier
    
    vulns_dict = {}
    
    subtree_depth = subtree_copy.depth()
    all_nodes = subtree_copy.all_nodes()
    # Analyze the (sub-)tree (e.g., tree of variable, entire CST/AST tree) bottom-up and use "taint propagation"
    # NOTE: This is a quite hackish way to approach bottom-up tree traversal. Now we KISS. We leave optimizations for later.
    for depth in range(subtree_depth, -1, -1):
        for node_tlib in all_nodes:
            if subtree_copy.level(node_tlib.identifier) == depth:
                node_antlr = node_tlib.data["node_antlr"]
                node_antlr_text = node_antlr.getText().encode('utf-8')
                node_state = node_tlib.data["node_state"]
                node_solved = False
                if node_state != None:
                    node_state.get("is_solved", False)

                if not node_solved and isinstance(node_antlr, LuaParser.StatContext):
                    children = list(node_antlr.getChildren())
                    if any([isinstance(c, LuaParser.NamelistContext) for c in children]) \
                    or any([isinstance(c, LuaParser.VarlistContext) for c in children]):
                        var_list = []
                        var_list_names = []

                        for c in children:
                            if isinstance(c, LuaParser.NamelistContext) \
                            or isinstance(c, LuaParser.VarlistContext):
                                var_list = list(c.getChildren())
                                var_list_names = [vartmp.getText().encode('utf-8') for vartmp in var_list]

                        for vartmp in var_list:
                            varnametmp = vartmp.getText().encode('utf-8')
                            
                            # Ignore variables separators, we will not be able to find them in variable maps/dicts and they do not have an expression (sub)tree associated to them from where to get propagate their value/taint
                            if varnametmp in [",", "="]:
                                continue

                            exptmp = variables_to_expressions_map.get(vartmp, None)
                            # TESTCASE: "cnum = cnum + 1" (from ntop-ng/host_details.lua)
                            # HACK: If we cannot find a variable name by node's object, we try to find it by name
                            if exptmp == None:
                                exptmp = variables_to_expressions_map_by_name.get(varnametmp, None)

                            vartmp_id = mslutils.get_object_id(vartmp)
                            # NOTE: exptmp is a tuple ("varname", "varnode", "expnode")
                            exptmp_id = mslutils.get_object_id(exptmp[2])

                            vartmp_tlib_node = cst_tlib[vartmp_id]
                            vartmp_tlib_node_state = vartmp_tlib_node.data["node_state"]
                            exptmp_tlib_node = cst_tlib[exptmp_id]
                            exptmp_tlib_node_state = exptmp_tlib_node.data["node_state"]

                            d_tmp = dict(commonconfig.config_entry)
                            # HACK: 
                            d_tmp["is_solved"] = True

                            if exptmp_tlib_node_state["is_tainted"] and exptmp_tlib_node_state["taint_types"] != set([]) and exptmp_tlib_node_state["taint_types"].intersection(exptmp_tlib_node_state["passthru"]).difference(exptmp_tlib_node_state["sanitize"]).union(exptmp_tlib_node_state["unsanitize"]) != set([]):
                                d_tmp["is_tainted"] = exptmp_tlib_node_state["is_tainted"]
                                d_tmp["taint_types"] = set(exptmp_tlib_node_state["taint_types"])
                                d_tmp["passthru"] = set(exptmp_tlib_node_state["passthru"])

                            node_solved = True
                            # HACK: 
                            d_tmp["is_solved"] = node_solved
                            # Remove existing state information from current node
                            del vartmp_tlib_node.data["node_state"]
                            # And update with the new state information
                            vartmp_tlib_node.data["node_state"] = d_tmp

                if not node_solved and isinstance(node_antlr, LuaParser.ExpContext):
                    children = list(node_antlr.getChildren())
                    # Concatenate operator
                    if any([isinstance(c, LuaParser.OperatorStrcatContext) for c in children]):
                        d_tmp = dict(commonconfig.config_entry)
                        for ch in children:
                            ch_node_id = mslutils.get_object_id(ch)
                            ch_node_tlib = subtree_copy[ch_node_id]
                            ch_node_state = ch_node_tlib.data["node_state"]
                            d_tmp["is_tainted"] = any([d_tmp["is_tainted"], ch_node_state["is_tainted"]])
                            d_tmp["passthru"] = commonconfig.DEFAULT_VULN_SET
                            d_tmp["taint_types"] = d_tmp["taint_types"].union(ch_node_state["taint_types"])

                        node_solved = True
                        # HACK: 
                        d_tmp["is_solved"] = node_solved
                        # Remove existing state information from current node
                        del node_tlib.data["node_state"]
                        # And update with the new state information
                        node_tlib.data["node_state"] = d_tmp

                if not node_solved and any([isinstance(node_antlr, LuaParser.PrefixexpContext), isinstance(node_antlr, LuaParser.FunctioncallContext)]):
                    children = list(node_antlr.getChildren())
                    # Function call
                    if len(children) == 2 and isinstance(children[0], LuaParser.VarOrExpContext) and isinstance(children[1], LuaParser.NameAndArgsContext):
                        ch0 = children[0] #node_antlr.getChild(0)
                        ch1 = children[1] #node_antlr.getChild(1)
                        ch0_node_id = mslutils.get_object_id(ch0)
                        ch1_node_id = mslutils.get_object_id(ch1)
                        ch0_node_tlib = subtree_copy[ch0_node_id]
                        ch1_node_tlib = subtree_copy[ch1_node_id]
                        ch0_node_state = ch0_node_tlib.data["node_state"]
                        ch1_node_state = ch1_node_tlib.data["node_state"]

                        # Vulnerability detection:
                        #   if tainted node (argument) reaches a sensitive node (function) both having common taint/sensitiveness attributes
                        vuln_channels = ch1_node_state["taint_types"].intersection(ch0_node_state["sensitive"])
                        if len(vuln_channels) > 0 and ch1_node_state["is_tainted"]:
                            vulns_dict[node_antlr] = (node_antlr, vuln_channels, [ch0, ch1, ])

                        # Taint propagation:
                        # Create new state information for current node
                        d_tmp = dict(commonconfig.config_entry)
                        d_tmp["taint_types"] = ch1_node_state["taint_types"].intersection(ch0_node_state["passthru"]).difference(ch0_node_state["sanitize"]).union(ch0_node_state["unsanitize"])
                        d_tmp["is_tainted"] = any([node_state["is_tainted"], ch0_node_tlib.data["node_state"]["is_tainted"], ch1_node_tlib.data["node_state"]["is_tainted"],])
                        d_tmp["passthru"] = set(ch0_node_state["passthru"])
                        d_tmp["sanitize"] = set(ch0_node_state["sanitize"])
                        d_tmp["unsanitize"] = set(ch0_node_state["unsanitize"])
                        d_tmp["sensitive"] = set(ch0_node_state["sensitive"])

                        node_solved = True
                        # HACK: 
                        d_tmp["is_solved"] = node_solved
                        # Remove existing state information from current node
                        del node_tlib.data["node_state"]
                        # And update with the new state information
                        node_tlib.data["node_state"] = d_tmp

                if not node_solved and isinstance(node_antlr, LuaParser.VarContext) and not node_tlib.data["node_state"].get("is_solved", False):
                    vartmp_solved = variables_to_expressions_map_by_name.get(node_antlr_text, None)
                    if vartmp_solved != None:
                        vartmp_solved_id = mslutils.get_object_id(vartmp_solved[1])
                        vartmp_solved_tlib_node = cst_tlib[vartmp_solved_id]
                        vartmp_solved_tlib_node_state = vartmp_solved_tlib_node.data["node_state"]

                        d_tmp = dict(vartmp_solved_tlib_node_state)

                        node_solved = True
                        # HACK: 
                        d_tmp["is_solved"] = node_solved
                        # Remove existing state information from current node
                        del node_tlib.data["node_state"]
                        # And update with the new state information
                        node_tlib.data["node_state"] = d_tmp

                if not node_solved:
                    is_tainted = False #node_state["is_tainted"]
                    taint_types = node_state["taint_types"]
                    passthru = node_state["passthru"]
                    for node_tlib_child in subtree_copy.children(node_tlib.identifier):
                        node_tlib_child_state = node_tlib_child.data["node_state"]
                        is_tainted = is_tainted or node_tlib_child_state["is_tainted"]
                        #taint_types = taint_types or node_tlib_child_state["taint_types"]
                        taint_types = taint_types.union(node_tlib_child_state["taint_types"])
                        passthru = passthru.union(node_tlib_child_state["passthru"])

                    # Create new state information for current node
                    d_tmp = dict(commonconfig.config_entry)

                    # Logic to compute new state information
                    d_tmp["is_tainted"] = node_state["is_tainted"] or is_tainted
                    
                    cancelation = node_state["sanitize"].intersection(node_state["unsanitize"])
                    sanitize_real = node_state["sanitize"].difference(cancelation)
                    unsanitize_real = node_state["unsanitize"].difference(cancelation)
                    d_tmp["taint_types"] = taint_types.union(unsanitize_real).difference(sanitize_real)
                    d_tmp["passthru"] = node_state["passthru"]
                    d_tmp["passthru"] = passthru
                    d_tmp["sanitize"] = node_state["sanitize"]
                    d_tmp["unsanitize"] = node_state["unsanitize"]
                    # Sensitiveness of a node is not propagated, it's a node-specific property
                    d_tmp["sensitive"] = node_state["sensitive"]

                    node_solved = True
                    # HACK: 
                    d_tmp["is_solved"] = node_solved
                    # Remove existing state information from current node
                    del node_tlib.data["node_state"]
                    # And update with the new state information
                    node_tlib.data["node_state"] = d_tmp

    return (subtree_copy, c_tree_parent_parent_identifier, vulns_dict)


def solve_variables_dependency_and_order(cst_tlib, variables_all):
    variables_ordered = []
    variables_independent = set([])
    variables_dependent = {}
    variables_to_expressions_map = {}

    for v in variables_all:
        c = v["node"]
        c_text = v["text"]
        c_node_id = mslutils.get_object_id(c)
        c_tree_node = cst_tlib[c_node_id]
        c_tree_parent = cst_tlib.parent(c_node_id)
        c_tree_parent_parent = cst_tlib.parent(c_tree_parent.identifier)
        c_parent_parent = c_tree_parent_parent.data["node_antlr"]
        subtree_copy = Tree(cst_tlib.subtree(c_tree_parent_parent.identifier), deep = False)

        assert(isinstance(c_parent_parent, LuaParser.StatContext))
        assert(isinstance(v["node"], antlr4.tree.Tree.TerminalNodeImpl) or isinstance(v["node"], LuaParser.VarContext))
        assert(isinstance(v["parent"], LuaParser.NamelistContext) or isinstance(v["parent"], LuaParser.VarlistContext))

        var_list = []
        exp_list = []
        for ctmp in c_parent_parent.getChildren():
            if isinstance(ctmp, LuaParser.NamelistContext) or isinstance(ctmp, LuaParser.VarlistContext):
                var_list = list(ctmp.getChildren())
            if isinstance(ctmp, LuaParser.ExplistContext):
                exp_list = list(ctmp.getChildren())

        var_index = var_list.index(c) if c in var_list else -1
        exp_index = var_index

        assert(var_index in range(0, len(var_list)))

        # TESTCASE: "for id, name in pairs(stats) do" (from ntop-ng/host_details.lua)
        # NOTE: These assertions do not hold. Could implement if we add "sideeffect" annotations to statements such as "pairs(stats)". For now we KISS.
        #assert(len(var_list) == len(exp_list))
        #assert(exp_index in range(0, len(exp_list)))
        
        # TESTCASE: "for id, name in pairs(stats) do" (from ntop-ng/host_details.lua)
        # NOTE: For now we KISS.
        if exp_index > len(exp_list):
            exp_index = len(exp_list) - 1
        variables_to_expressions_map[var_list[var_index]] = (c_text, var_list[var_index], exp_list[exp_index])

        # First assume that currently analyzed variable is dependent but has an empty dependency set 
        variables_dependent[c_text] = set([])
        
        # Create a non-deep copy of the expression (sub)tree _which explicitly corresponds_ (important for multiple assignment/declaration statements) to the currently iterated variable. 
        subtree_copy_var_exp = Tree(cst_tlib.subtree(mslutils.get_object_id(exp_list[exp_index])), deep = False)
        for n in subtree_copy_var_exp.all_nodes():
            # If the subtree of the currently analyzed variable has VarContext nodes (including itself)
            if isinstance(n.data["node_antlr"], LuaParser.VarContext):
                n_text = n.data["node_antlr"].getText().encode('utf-8')

                # And there is a sub-node (different from the variable itself) that is in the variables list
                if (n_text != c_text) and n_text in [vtmp["text"] for vtmp in variables_all]:
                    # Then assume it is a local/global variable dependency and add it to the set of dependencies
                    variables_dependent[c_text] = variables_dependent[c_text].union([n_text])

        # If the set of dependencies of currently analyzed variable is indeed empty
        if len(variables_dependent[c_text]) == 0:
            # Remove this variable from the dictionary of dependent variables
            variables_dependent.pop(c_text)
            # Add this variable to the set of independent variables
            variables_independent = variables_independent.union([c_text])

    # Variables need to be "solved" (for tainting) in order, start with independent variables
    variables_ordered = [v for v in variables_independent]

    # Here we decide the order of precedence/dependency in which to solve the variables, so that TAINT propagates according to dependency
    # WARN: Possible deadlock in case of any BUG: below
    while True:
        variables_ordered_tmp = list(variables_ordered)
        # Take each other variable that has dependencies
        for (v, s) in variables_dependent.items():
            # Check if their set of dependencies are already in the ordered solving-list
            if s.issubset(variables_independent):
                variables_ordered.append(v)
                variables_independent = variables_independent.union([v])
                variables_dependent.pop(v)

        # If there are no more variables with dependencies, we are done with variable ordering based on dependencies
        if len(variables_dependent) == 0:
            break

        if len(set(variables_ordered).difference(set(variables_ordered_tmp))) == 0:
            print "BUG: Deadlock in variables order and dependency"
            # NOTE: TODO: We leave the break now as we want to KISS and not break the subsequent existing functionality.
            # raise BugException
            break

    return (variables_ordered, variables_independent, variables_dependent, variables_to_expressions_map)


