#!/usr/bin/env python
# -*- coding: utf-8 -*-

## MoonShineLua (MSL)
##
## Copyright 2016 Andrei Costin (andrei@firmware.re) for Firmware.RE
## with support from Tjaldur Software Governance Solutions and NLnet.nl.
##
## See LICENSE file for details.


'''
Unit tests for MSL.
'''

import os
import sys
import pprint
import unittest
import msl.tests.common as commontests
import msl.utils.lua_parser_json as mslparser
import msl.utils.utils as mslutils
import msl.taint.utils as taintutils

class LangSec17Tests(unittest.TestCase):
    ############################################################################
    def setUp(self):
        (self.dir, self.dir_realpath) = commontests.get_tests_dir_prefix()
        self.code = ""
        self.clsname = self.__class__.__name__
        self.funcname = sys._getframe().f_code.co_name

    ############################################################################
    def tearDown(self):
        self.dir = None
        self.dir_realpath = None
        self.code = ""
        self.clsname = None
        self.funcname = None

    ############################################################################
    def test_Listing7(self):
        self.funcname = sys._getframe().f_code.co_name

        self.code = \
"""
local luatech = 'CGILua'
local username = cgilua.QUERY.username or 'Unknown username'

cgilua.htmlheader()
cgilua.put([[
<title>XSS Lua ]] .. luatech .. [[</title>
]])
cgilua.put([[ Hello, ]] .. username)
"""

        assert(self.funcname != None)
        assert(self.code != None)

        tmp_fname = commontests.dump_test_code(self.dir_realpath, self.clsname, self.funcname, self.code)
        assert(tmp_fname != None)

        luajsonres = mslparser.lua_get_json(tmp_fname)
        assert(luajsonres != None)

        cst_tlib = luajsonres["cst_tlib"]
        assert(cst_tlib != None)
        
        variables_all = luajsonres["variables_def"] + luajsonres["variables_local_def"]
        (variables_ordered, variables_independent, variables_dependent, variables_to_expressions_map) = taintutils.solve_variables_dependency_and_order(cst_tlib, variables_all)

        vulns_dict_all = {}

        ########################################################################
        
        for vname in variables_ordered:
            # Get the JSON object related to this variable
            v = [d for d in variables_all if d["text"] == vname][0]
            
            # Solve/taint current variable's (sub)tree
            (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, v["node"], v["text"], variables_to_expressions_map)
            
            # Merge current variable vulnerabilities dict into the all vulnerabilities dict
            vulns_dict_all.update(vulns_dict_var)
            
        ########################################################################
        
        cst_tlib_depth = cst_tlib.depth()
        all_nodes = cst_tlib.all_nodes()
        # We go just one level lower than the root node, therefore the second parameter is 0 instead of -1
        for depth in range(cst_tlib_depth, 0, -1):
            for node_tlib in all_nodes:
                node_solved = node_tlib.data.get("node_state", {"is_solved": False}).get("is_solved", False)
                if cst_tlib.level(node_tlib.identifier) == depth and not node_solved:
                    node_antlr = node_tlib.data["node_antlr"]
                    node_antlr_text = node_antlr.getText().encode('utf-8')
                    
                    # Solve/taint current unsolved node (sub)tree
                    (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, node_antlr, node_antlr_text, variables_to_expressions_map)
                    
                    # Merge current variable vulnerabilities dict into the all vulnerabilities dict
                    vulns_dict_all.update(vulns_dict_var)
                    
        ########################################################################

        vuln_types = set([])
        vuln_lines = set([])
        vuln_columns = set([])

        for vuln in vulns_dict_all.values():
            node_antlr = vuln[0]
            node_antlr_text = node_antlr.getText().encode('utf-8')
            type_of_vuln = vuln[1]

            vuln_types = vuln_types.union(type_of_vuln)
            vuln_lines = vuln_lines.union([node_antlr.start.line])
            vuln_columns = vuln_columns.union([node_antlr.start.column])

            # DEBUG:
            #print "VULN type(s):", type_of_vuln
            #print "VULN line:", node_antlr.start.line
            #print "VULN column:", node_antlr.start.column
            #print "VULN code:", node_antlr_text.replace("\n", "").replace("\r", "").replace("-", "").replace('"', '')
            #print

        # Check variable parsing and detection
        assert(len(variables_all) == 2)
        assert(len(variables_ordered) == 2)
        assert(set(variables_ordered) == set(["luatech", "username"]))

        # Check vulnerability detection, types and locations
        assert(len(vulns_dict_all) == 1)
        assert(vuln_types == set(["xss"]))
        assert(vuln_lines == set([9]))
        assert(vuln_columns == set([0]))

    ############################################################################
    def test_Listing8(self):
        self.funcname = sys._getframe().f_code.co_name

        self.code = \
"""
local luatech = 'CGIlua'
local username = cgilua.QUERY.username or 'root'

cgilua.htmlheader()

cgilua.put([[
<title>OS Command Injection Lua ]] .. luatech .. [[</title>
]])

local handle = io.popen("ls -lart /home/" .. username)
local data = handle:read("*a")
handle:close()

cgilua.put([[
Listing home dir: 
]] .. data)
"""

        assert(self.funcname != None)
        assert(self.code != None)

        tmp_fname = commontests.dump_test_code(self.dir_realpath, self.clsname, self.funcname, self.code)
        assert(tmp_fname != None)

        luajsonres = mslparser.lua_get_json(tmp_fname)
        assert(luajsonres != None)

        cst_tlib = luajsonres["cst_tlib"]
        assert(cst_tlib != None)
        
        variables_all = luajsonres["variables_def"] + luajsonres["variables_local_def"]
        (variables_ordered, variables_independent, variables_dependent, variables_to_expressions_map) = taintutils.solve_variables_dependency_and_order(cst_tlib, variables_all)

        vulns_dict_all = {}

        ########################################################################
        
        for vname in variables_ordered:
            # Get the JSON object related to this variable
            v = [d for d in variables_all if d["text"] == vname][0]
            
            # Solve/taint current variable's (sub)tree
            (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, v["node"], v["text"], variables_to_expressions_map)
            
            # Merge current variable vulnerabilities dict into the all vulnerabilities dict
            vulns_dict_all.update(vulns_dict_var)
            
        ########################################################################
        
        cst_tlib_depth = cst_tlib.depth()
        all_nodes = cst_tlib.all_nodes()
        # We go just one level lower than the root node, therefore the second parameter is 0 instead of -1
        for depth in range(cst_tlib_depth, 0, -1):
            for node_tlib in all_nodes:
                node_solved = node_tlib.data.get("node_state", {"is_solved": False}).get("is_solved", False)
                if cst_tlib.level(node_tlib.identifier) == depth and not node_solved:
                    node_antlr = node_tlib.data["node_antlr"]
                    node_antlr_text = node_antlr.getText().encode('utf-8')
                    
                    # Solve/taint current unsolved node (sub)tree
                    (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, node_antlr, node_antlr_text, variables_to_expressions_map)
                    
                    # Merge current variable vulnerabilities dict into the all vulnerabilities dict
                    vulns_dict_all.update(vulns_dict_var)
                    
        ########################################################################

        vuln_types = set([])
        vuln_lines = set([])
        vuln_columns = set([])

        for vuln in vulns_dict_all.values():
            node_antlr = vuln[0]
            node_antlr_text = node_antlr.getText().encode('utf-8')
            type_of_vuln = vuln[1]

            vuln_types = vuln_types.union(type_of_vuln)
            vuln_lines = vuln_lines.union([node_antlr.start.line])
            vuln_columns = vuln_columns.union([node_antlr.start.column])

            # DEBUG:
            #print "VULN type(s):", type_of_vuln
            #print "VULN line:", node_antlr.start.line
            #print "VULN column:", node_antlr.start.column
            #print "VULN code:", node_antlr_text.replace("\n", "").replace("\r", "").replace("-", "").replace('"', '')
            #print

        # Check variable parsing and detection
        assert(len(variables_all) == 4)
        assert(len(variables_ordered) == 4)
        assert(set(variables_ordered) == set(["luatech", "username", "handle", "data"]))

        # Check vulnerability detection, types and locations
        assert(len(vulns_dict_all) == 1)
        assert(vuln_types == set(["cmdi"]))
        assert(vuln_lines == set([11]))
        assert(vuln_columns == set([15]))

    ############################################################################
    def test_Listing9_Listing10(self):
        self.funcname = sys._getframe().f_code.co_name

        self.code = \
"""
local luatech = 'CGIlua'
local textfile = cgilua.QUERY.textfile or 'README'

local f = io.open(textfile .. ".txt")
local data = f:read("*a")
f:close()

cgilua.htmlheader()

cgilua.put([[
<title>Null-Byte Injection Lua ]] .. luatech .. [[</title>
]])

cgilua.put([[
Your text file contains:
]] .. data)
"""

        assert(self.funcname != None)
        assert(self.code != None)

        tmp_fname = commontests.dump_test_code(self.dir_realpath, self.clsname, self.funcname, self.code)
        assert(tmp_fname != None)

        luajsonres = mslparser.lua_get_json(tmp_fname)
        assert(luajsonres != None)

        cst_tlib = luajsonres["cst_tlib"]
        assert(cst_tlib != None)
        
        variables_all = luajsonres["variables_def"] + luajsonres["variables_local_def"]
        (variables_ordered, variables_independent, variables_dependent, variables_to_expressions_map) = taintutils.solve_variables_dependency_and_order(cst_tlib, variables_all)

        vulns_dict_all = {}

        ########################################################################
        
        for vname in variables_ordered:
            # Get the JSON object related to this variable
            v = [d for d in variables_all if d["text"] == vname][0]
            
            # Solve/taint current variable's (sub)tree
            (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, v["node"], v["text"], variables_to_expressions_map)
            
            # Merge current variable vulnerabilities dict into the all vulnerabilities dict
            vulns_dict_all.update(vulns_dict_var)
            
        ########################################################################
        
        cst_tlib_depth = cst_tlib.depth()
        all_nodes = cst_tlib.all_nodes()
        # We go just one level lower than the root node, therefore the second parameter is 0 instead of -1
        for depth in range(cst_tlib_depth, 0, -1):
            for node_tlib in all_nodes:
                node_solved = node_tlib.data.get("node_state", {"is_solved": False}).get("is_solved", False)
                if cst_tlib.level(node_tlib.identifier) == depth and not node_solved:
                    node_antlr = node_tlib.data["node_antlr"]
                    node_antlr_text = node_antlr.getText().encode('utf-8')
                    
                    # Solve/taint current unsolved node (sub)tree
                    (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, node_antlr, node_antlr_text, variables_to_expressions_map)
                    
                    # Merge current variable vulnerabilities dict into the all vulnerabilities dict
                    vulns_dict_all.update(vulns_dict_var)
                    
        ########################################################################

        vuln_types = set([])
        vuln_lines = set([])
        vuln_columns = set([])

        for vuln in vulns_dict_all.values():
            node_antlr = vuln[0]
            node_antlr_text = node_antlr.getText().encode('utf-8')
            type_of_vuln = vuln[1]

            vuln_types = vuln_types.union(type_of_vuln)
            vuln_lines = vuln_lines.union([node_antlr.start.line])
            vuln_columns = vuln_columns.union([node_antlr.start.column])

            # DEBUG:
            #print "VULN type(s):", type_of_vuln
            #print "VULN line:", node_antlr.start.line
            #print "VULN column:", node_antlr.start.column
            #print "VULN code:", node_antlr_text.replace("\n", "").replace("\r", "").replace("-", "").replace('"', '')
            #print

        # Check variable parsing and detection
        assert(len(variables_all) == 4)
        assert(len(variables_ordered) == 4)
        assert(set(variables_ordered) == set(["luatech", "textfile", "f", "data"]))

        # Check vulnerability detection, types and locations
        assert(len(vulns_dict_all) == 1)
        assert(vuln_types == set(["pathtrav", "nullbyte"]))
        assert(vuln_lines == set([5]))
        assert(vuln_columns == set([10]))

    ############################################################################
    def test_Listing11(self):
        self.funcname = sys._getframe().f_code.co_name

        self.code = \
"""
luatech = 'CGIlua'
local template = cgilua.QUERY.template or "default_template"

cgilua.htmlheader()
cgilua.handlelp(template .. ".lp")
"""

        assert(self.funcname != None)
        assert(self.code != None)

        tmp_fname = commontests.dump_test_code(self.dir_realpath, self.clsname, self.funcname, self.code)
        assert(tmp_fname != None)

        luajsonres = mslparser.lua_get_json(tmp_fname)
        assert(luajsonres != None)

        cst_tlib = luajsonres["cst_tlib"]
        assert(cst_tlib != None)
        
        variables_all = luajsonres["variables_def"] + luajsonres["variables_local_def"]
        (variables_ordered, variables_independent, variables_dependent, variables_to_expressions_map) = taintutils.solve_variables_dependency_and_order(cst_tlib, variables_all)

        vulns_dict_all = {}

        ########################################################################
        
        for vname in variables_ordered:
            # Get the JSON object related to this variable
            v = [d for d in variables_all if d["text"] == vname][0]
            
            # Solve/taint current variable's (sub)tree
            (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, v["node"], v["text"], variables_to_expressions_map)
            
            # Merge current variable vulnerabilities dict into the all vulnerabilities dict
            vulns_dict_all.update(vulns_dict_var)
            
        ########################################################################
        
        cst_tlib_depth = cst_tlib.depth()
        all_nodes = cst_tlib.all_nodes()
        # We go just one level lower than the root node, therefore the second parameter is 0 instead of -1
        for depth in range(cst_tlib_depth, 0, -1):
            for node_tlib in all_nodes:
                node_solved = node_tlib.data.get("node_state", {"is_solved": False}).get("is_solved", False)
                if cst_tlib.level(node_tlib.identifier) == depth and not node_solved:
                    node_antlr = node_tlib.data["node_antlr"]
                    node_antlr_text = node_antlr.getText().encode('utf-8')
                    
                    # Solve/taint current unsolved node (sub)tree
                    (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, node_antlr, node_antlr_text, variables_to_expressions_map)
                    
                    # Merge current variable vulnerabilities dict into the all vulnerabilities dict
                    vulns_dict_all.update(vulns_dict_var)
                    
        ########################################################################

        vuln_types = set([])
        vuln_lines = set([])
        vuln_columns = set([])

        for vuln in vulns_dict_all.values():
            node_antlr = vuln[0]
            node_antlr_text = node_antlr.getText().encode('utf-8')
            type_of_vuln = vuln[1]

            vuln_types = vuln_types.union(type_of_vuln)
            vuln_lines = vuln_lines.union([node_antlr.start.line])
            vuln_columns = vuln_columns.union([node_antlr.start.column])

            # DEBUG:
            #print "VULN type(s):", type_of_vuln
            #print "VULN line:", node_antlr.start.line
            #print "VULN column:", node_antlr.start.column
            #print "VULN code:", node_antlr_text.replace("\n", "").replace("\r", "").replace("-", "").replace('"', '')
            #print

        # Check variable parsing and detection
        assert(len(variables_all) == 2)
        assert(len(variables_ordered) == 2)
        assert(set(variables_ordered) == set(["luatech", "template"]))

        # Check vulnerability detection, types and locations
        assert(len(vulns_dict_all) == 1)
        assert(vuln_types == set(["lfi", "pathtrav", "nullbyte"]))
        assert(vuln_lines == set([6]))
        assert(vuln_columns == set([0]))

    ############################################################################
    def test_Listing12(self):
        self.funcname = sys._getframe().f_code.co_name

        self.code = \
"""
local luatech = 'CGIlua'
local u = cgilua.QUERY.username or ''
local p = cgilua.QUERY.password or ''
local logged = 0

local mysql = require "luasql.mysql"
local env  = mysql.mysql()
local conn = env:connect('mysqldb', 'mysqluser', 'mysqlpass')

cur, err = execute("select * from users where 
    username = '" .. u .. "' and password = '" .. p .. "'")

row = cur:fetch ({}, "a")

while row do
    row = cursor:fetch (row, "a")
    logged = 1
end

cgilua.htmlheader()

cgilua.put([[
<title>SQL Injection Lua ]] .. luatech .. [[</title>
]])

if logged == 1
then
    print("Yes")
    cgilua.put([[You successfully logged in as ]] .. u)
end

cursor:close()
conn:close()
env:close()
"""

        assert(self.funcname != None)
        assert(self.code != None)

        tmp_fname = commontests.dump_test_code(self.dir_realpath, self.clsname, self.funcname, self.code)
        assert(tmp_fname != None)

        luajsonres = mslparser.lua_get_json(tmp_fname)
        assert(luajsonres != None)

        cst_tlib = luajsonres["cst_tlib"]
        assert(cst_tlib != None)
        
        variables_all = luajsonres["variables_def"] + luajsonres["variables_local_def"]
        (variables_ordered, variables_independent, variables_dependent, variables_to_expressions_map) = taintutils.solve_variables_dependency_and_order(cst_tlib, variables_all)

        vulns_dict_all = {}

        ########################################################################
        
        for vname in variables_ordered:
            # Get the JSON object related to this variable
            v = [d for d in variables_all if d["text"] == vname][0]
            
            # Solve/taint current variable's (sub)tree
            (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, v["node"], v["text"], variables_to_expressions_map)
            
            # Merge current variable vulnerabilities dict into the all vulnerabilities dict
            vulns_dict_all.update(vulns_dict_var)
            
        ########################################################################
        
        cst_tlib_depth = cst_tlib.depth()
        all_nodes = cst_tlib.all_nodes()
        # We go just one level lower than the root node, therefore the second parameter is 0 instead of -1
        for depth in range(cst_tlib_depth, 0, -1):
            for node_tlib in all_nodes:
                node_solved = node_tlib.data.get("node_state", {"is_solved": False}).get("is_solved", False)
                if cst_tlib.level(node_tlib.identifier) == depth and not node_solved:
                    node_antlr = node_tlib.data["node_antlr"]
                    node_antlr_text = node_antlr.getText().encode('utf-8')
                    
                    # Solve/taint current unsolved node (sub)tree
                    (subtree_copy, subtree_copy_root_id, vulns_dict_var) = taintutils.solve_tainting(cst_tlib, node_antlr, node_antlr_text, variables_to_expressions_map)
                    
                    # Merge current variable vulnerabilities dict into the all vulnerabilities dict
                    vulns_dict_all.update(vulns_dict_var)
                    
        ########################################################################

        vuln_types = set([])
        vuln_lines = set([])
        vuln_columns = set([])

        for vuln in vulns_dict_all.values():
            node_antlr = vuln[0]
            node_antlr_text = node_antlr.getText().encode('utf-8')
            type_of_vuln = vuln[1]

            vuln_types = vuln_types.union(type_of_vuln)
            vuln_lines = vuln_lines.union([node_antlr.start.line])
            vuln_columns = vuln_columns.union([node_antlr.start.column])

            # DEBUG:
            #print "VULN type(s):", type_of_vuln
            #print "VULN line:", node_antlr.start.line
            #print "VULN column:", node_antlr.start.column
            #print "VULN code:", node_antlr_text.replace("\n", "").replace("\r", "").replace("-", "").replace('"', '')
            #print

        # Check variable parsing and detection
        assert(len(variables_ordered) == 10)
        assert(set(variables_ordered) == set(['p', 'logged', 'luatech', 'u', 'env', 'mysql', 'conn', 'cur', 'err', 'row']))

        # Check vulnerability detection, types and locations
        assert(len(vulns_dict_all) == 2)
        assert(vuln_types == set(["sqli", "xss"]))
        assert(vuln_lines == set([11, 30]))
        assert(vuln_columns == set([11, 4]))


def suite():
    suites = [LangSec17Tests, ]
    suite = unittest.TestSuite()
    for s in suites:
        suite.addTest(unittest.makeSuite(s))
    return suite


if __name__ == '__main__':
    unittest.main(defaultTest='suite')


