#!/usr/bin/env python
# -*- coding: utf-8 -*-

## MoonShineLua (MSL)
##
## Copyright 2016 Andrei Costin (andrei@firmware.re) for Firmware.RE
## with support from Tjaldur Software Governance Solutions and NLnet.nl.
##
## See LICENSE file for details.

'''
Program to parse a given Lua file and provide structured access to 
comments, strings, functions, variables, labels.

Basically it acts as a combination of "ctags" (exuberant-ctags) and 
"xgettext", but in a more versatile way focused on Lua language.
For example, "ctags" supports the extractions of only functions from Lua files.

Mainly, this program is intended to be imported/used by Binary Analysis Toolkit:
$BAT/src/maintenance/createdb.py
https://raw.githubusercontent.com/armijnhemel/binaryanalysis/master/src/maintenance/createdb.py

This program can also be used in standalone mode for self-tests, 
or in a command-line/external call setting.
'''

import antlr4
from antlr4 import *

from msl.parsing.lua.LuaLexer import LuaLexer
from msl.parsing.lua.LuaParser import LuaParser
#from msl.parsing.lua.LuaPrettyPrinter import LuaPrettyPrinter
import msl.utils.encodings as mslencodings
import msl.utils.utils as mslutils
import msl.taint.utils as taintutils
import msl.config.common as commonconfig
import msl.config.defaultconfig as defaultconfig

import sys
import os
import pprint
import io

def build_tree_extended(t, parent_id, children_nodes):
    exceptions = []
    strings = []
    labels_def = []
    labels_use = []
    functions_def = []
    variables_def = []
    variables_local_def = []

    for c in children_nodes:
        try:
            ####################################################################
            # Static analysis and taint propagation related functionality

            c_node_id = mslutils.get_object_id(c)
            c_text_tmp = c.getText().encode('utf-8')
            c_node_text = str(type(c)).replace("antlr4.tree.Tree.", "").replace("msl.parsing.lua.LuaParser.", "").replace("class ", "")
            try:
                c_node_text = c_node_text + " " + c_node_id + " " + c_text_tmp.replace("\n", "").replace("\r", "").replace("-", "").replace('"', '')[0:50]
            except Exception, ex:
                ex = ex

            c_state = commonconfig.get_security_attributes(c_text_tmp, defaultconfig.config_db)

            t.create_node(c_node_text, c_node_id, parent = parent_id, data = {"node_antlr": c, "node_state": c_state})

            ####################################################################
            # BAT related functionality
            
            if isinstance(c, LuaParser.StringContext):
                # STRINGS
                strings.append({"text": c.getChild(0).getText().encode('utf-8'), "line": c.start.line, "column": c.start.column, "node": c.getChild(0), "parent": c})


            if isinstance(c, LuaParser.LabelContext):
                # LABELS_DEF
                labels_def.append({"text": c.getChild(1).getText().encode('utf-8'), "line": c.start.line, "column": c.start.column, "node": c.getChild(1), "parent": c})


            if isinstance(c, LuaParser.VarlistContext):
                # VARIABLES_DEF
                for i in range(0, c.getChildCount()):
                    if i % 2 == 0:
                        variables_def.append({"text": c.getChild(i).getText().encode('utf-8'), "line": c.start.line, "column": c.start.column, "node": c.getChild(i), "parent": c})


            if isinstance(c, LuaParser.StatContext):
                # LABELS_USE
                if c.getChild(0).getText().encode('utf-8') == 'goto':
                    labels_use.append({"text": c.getChild(1).getText().encode('utf-8'), "line": c.start.line, "column": c.start.column, "node": c.getChild(1), "parent": c})

                if c.getChild(0).getText().encode('utf-8') == 'function':
                    # FUNCTIONS_DEF
                    functions_def.append({"text": c.getChild(1).getText().encode('utf-8'), "line": c.start.line, "column": c.start.column, "node": c.getChild(1), "parent": c})

                if c.getChild(0).getText().encode('utf-8') == 'local':
                    # FUNCTIONS_DEF
                    if c.getChild(1).getText().encode('utf-8') == 'function':
                        functions_def.append({"text": c.getChild(2).getText().encode('utf-8'), "line": c.start.line, "column": c.start.column, "node": c.getChild(2), "parent": c})
                    # VARIABLES_LOCAL_DEF
                    elif isinstance(c.getChild(1), LuaParser.NamelistContext):
                        namelist = c.getChild(1)
                        for i in range(0, namelist.getChildCount()):
                            if i % 2 == 0:
                                variables_local_def.append({"text": namelist.getChild(i).getText().encode('utf-8'), "line": namelist.start.line, "column": namelist.start.column, "node": namelist.getChild(i), "parent": namelist})


                if c.getChild(0).getText().encode('utf-8') == 'for':
                    # VARIABLES_DEF
                    if isinstance(c.getChild(1), antlr4.tree.Tree.TerminalNodeImpl):
                        variables_def.append({"text": c.getChild(1).getText().encode('utf-8'), "line": c.start.line, "column": c.start.column, "node": c.getChild(1), "parent": c})
                    # VARIABLES_DEF
                    elif isinstance(c.getChild(1), LuaParser.NamelistContext):
                        namelist = c.getChild(1)
                        for i in range(0, namelist.getChildCount()):
                            if i % 2 == 0:
                                variables_def.append({"text": namelist.getChild(i).getText().encode('utf-8'), "line": namelist.start.line, "column": namelist.start.column, "node": namelist.getChild(i), "parent": namelist})

            ####################################################################

            # Recursively explore depth-first style
            if c.getChildCount() > 0:
                (tmp_exceptions, tmp_strings, tmp_labels_def, tmp_labels_use, tmp_functions_def, tmp_variables_def, tmp_variables_local_def) = build_tree_extended(t, c_node_id, c.getChildren())
                exceptions += tmp_exceptions
                strings += tmp_strings
                labels_def += tmp_labels_def
                labels_use += tmp_labels_use
                functions_def += tmp_functions_def
                variables_def += tmp_variables_def
                variables_local_def += tmp_variables_local_def

        except Exception, ex:
            exceptions.append({"node": c, "exception": ex})
            pass

    return (exceptions, strings, labels_def, labels_use, functions_def, variables_def, variables_local_def)


def lua_get_parse_components(filename, encodings_try = ['utf-8']):
    """Parses a Lua script file, and returns the ANTLR4 parser internals. 
    
    Parses a Lua script file, trying the best encoding that can parse the 
    supplied file, and gets the main ANTLR4 internal components 
    (such as parser, lexer, CST parse tree, comments). 
    
    Parameters
    ----------
    filename : str
               Lua script filename to parse.
               
    encodings_try : list of str
                    List of encodings to be tried when reading and parsing the 
                    Lua script file.

    Returns
    -------
    res : tuple of objects or None
          Tuple of (lexer, parser, cst_parse_tree, comments, encoding, exceptions), 
          or None on error.
    """

    input = None
    lexer = None
    parser = None
    cst_parse_tree = None
    comments = []
    exceptions = []
    
    if encodings_try == None or len(encodings_try) == 0:
        encodings_try = ['utf-8']

    for encoding in encodings_try:
        try:
            input = FileStream(fileName = filename, encoding = encoding)
            
            lexer = LuaLexer(input)

            # https://github.com/antlr/antlr4/blob/47e268dfeaaf0f08930c7d6428b57ee6a5f8c44d/runtime/Python2/src/antlr4/CommonTokenStream.py
            channel = Token.DEFAULT_CHANNEL
            stream = CommonTokenStream(lexer, channel = channel)

            # NOTE: This block of code is intended ONLY to extract comment related tokens, i.e., using Token.HIDDEN_CHANNEL
            start_index = 0
            while True:
                token_index = stream.nextTokenOnChannel(start_index, channel = Token.HIDDEN_CHANNEL)
                if token_index == -1:
                    break
                token = stream.get(token_index)
                # https://github.com/antlr/antlr4/blob/47e268dfeaaf0f08930c7d6428b57ee6a5f8c44d/runtime/Python2/src/antlr4/CommonTokenFactory.py
                comments.append({"text": token.text.encode('utf-8'), "line": token.line, "column": token.column, "node": token, "parent": None})
                start_index = token_index+1
            
            parser = LuaParser(stream)

            # In "Lua5.2.g4", the first grammar rule is "chunk", therefore we get the CST Parse Tree from parser using parser.chunk()
            cst_parse_tree = parser.chunk()
            
            return (lexer, parser, cst_parse_tree, comments, encoding, [])

        except Exception, ex:
            exceptions.append(ex)
            pass

    return (None, None, None, None, None, exceptions)


def lua_get_json(filename, encodings_try = ['utf-8']):
    """Parses a Lua script file, and returns corresponding JSON. 
    
    Parses a Lua script file, extracts Lua entities: strings, functions, 
    variables, labels, and returns them to the caller in a structured format 
    (e.g., dictionary/JSON) for easier post-processing.
    
    Parameters
    ----------
    filename : str
               Lua script filename to parse.
               
    encodings_try : list of str
                    List of encodings to be tried when reading and parsing the 
                    Lua script file.

    Returns
    -------
    luajsonres : dictionary or None
                 Dictionary with meta file information and all extracted 
                 entities from Lua script file, or None on error.
    """
    
    # Basic arguments checks
    if filename == None or not os.path.isfile(filename):
        return None

    comments = []
    exceptions_parse = []
    exceptions_walk = []
    strings = []
    labels_def = []
    labels_use = []
    functions_def = []
    variables_def = []
    variables_local_def = []
    # CST (Concrete Syntax Tree) aka Parse Tree
    cst_parse_tree = None
    encoding_parse = None
    
    luajsonres = None

    (lexer, parser, cst_parse_tree, comments, encoding_parse, exceptions_parse) = lua_get_parse_components(filename, encodings_try)
    
    if exceptions_parse != None and len(exceptions_parse) > 0:
        luajsonres = {}
        luajsonres["exceptions_parse"] = exceptions_parse
    
    # We got a good CST Parse Tree, so we process it
    if cst_parse_tree != None:
        variables_def_dupl = []

        from treelib import Tree, Node
        from treelib.tree import NodeIDAbsentError
        
        cst_parse_tree_id = mslutils.get_object_id(cst_parse_tree)
        cst_tlib = Tree()
        cst_tlib.create_node("cst_parse_tree", cst_parse_tree_id, data = {"node_antlr": cst_parse_tree, "node_state": dict(commonconfig.config_entry)})

        (exceptions_walk, strings, labels_def, labels_use, functions_def, variables_def_dupl, variables_local_def) = build_tree_extended(cst_tlib, cst_parse_tree_id, cst_parse_tree.getChildren())

        # Deduplicate the encountered variables which are not explicitly local, keeping just first place of occurrence
        for var in variables_def_dupl:
            # NOTE: FIXME: Untrivial to properly do without proper semantic analysis/walking of CST/AST. For now we KISS.
            # NOTE: "ctags" does not deduplicate global variables with the same name
            # NOTE: "ctags" does not print local variables
            # http://stackoverflow.com/a/7271523
            already_indexed = [entry["text"] for entry in variables_def if "text" in entry.keys()]
            if not var.get("text") in already_indexed:
                variables_def.append(var)

        ########################################################################
        # Solving variables inter-dependencies and order for taint analysis
        
        variables_ordered = []
        variables_independent = set([])
        variables_dependent = {}
        variables_to_expressions_map = {}
        
        variables_all = variables_def + variables_local_def
        (variables_ordered, variables_independent, variables_dependent, variables_to_expressions_map) = taintutils.solve_variables_dependency_and_order(cst_tlib, variables_all)
        
        ########################################################################
        # Prepare the result, ready to be consumed as dict/JSON by the callers

        luajsonres = {}
        luajsonres["file_path"] = filename
        luajsonres["file_name"] = os.path.basename(filename)
        luajsonres["file_realpath"] = os.path.realpath(filename)
        luajsonres["file_realname"] = os.path.basename(luajsonres["file_realpath"])
        luajsonres["file_encoding"] = encoding_parse
        luajsonres["comments"] = comments
        luajsonres["strings"] = strings
        luajsonres["labels_def"] = labels_def
        luajsonres["labels_use"] = labels_use
        luajsonres["functions_def"] = functions_def
        luajsonres["variables_def"] = variables_def
        luajsonres["variables_local_def"] = variables_local_def
        luajsonres["exceptions_parse"] = exceptions_parse
        luajsonres["exceptions_walk"] = exceptions_walk
        luajsonres["variables_ordered"] = variables_ordered
        luajsonres["variables_independent"] = variables_independent
        luajsonres["variables_dependent"] = variables_dependent
        luajsonres["cst_parse_tree"] = cst_parse_tree
        luajsonres["cst_tlib"] = cst_tlib

        return luajsonres

        # NOTE: if we want to walk the CST Parse Tree in an ANTLR4 style, use the ParseTreeWalker and code the CST-to-AST tree "folding" logic into LuaPrettyPrinter. For now we KISS.
        #printer = LuaPrettyPrinter()
        #walker = ParseTreeWalker()
        #walker.walk(printer, cst_parse_tree)

    return luajsonres


if __name__ == '__main__':
    luajsonres = None

    for f in sys.argv[1:]:
        luajsonres = {}

        print f

        try:
            # Iterate the encodings to try on the file by using guessed or most-likely to appear encodings
            #encodings_try = [encodings.guess_file_encoding(f).get('encoding', 'utf-8')] + encodings.list_encodings_top() + encodings.list_encodings_all()
            encodings_try = [mslencodings.guess_file_encoding(f).get('encoding', 'utf-8')] + mslencodings.list_encodings_top() + mslencodings.list_encodings_all()
            #luajsonres = lua_get_json(f), encodings_try)
            luajsonres = lua_get_json(f)#, encodings_try)
        except Exception, ex:
            print ex

        print
        pprint.pprint(luajsonres)
        print


